@@ -10,11 +10,14 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 from __future__ import unicode_literals
-from spotpy import database, objectivefunctions
+from spotpy import database
 from spotpy import parameter
 import numpy as np
 import time
 import threading
+import inspect
+import sys
+
 
 try:
     from queue import Queue
@@ -36,35 +39,67 @@ class _RunStatistic(object):
     Usage:
     status = _RunStatistic()
     status(rep,like,params)
-
     """
 
-    def __init__(self):
+    def __init__(self, repetitions, algorithm_name, optimization_direction, parnames):
+        self.optimization_direction = optimization_direction #grid, mazimize, minimize
+        print('Initializing the ',algorithm_name,' with ',repetitions,' repetitions')
+        if optimization_direction == 'minimize':
+            self.compare = self.minimizer
+            print('The objective function will be minimized')
+        if optimization_direction == 'maximize':
+            self.compare = self.maximizer
+            print('The objective function will be minimized')
+        if optimization_direction == 'grid':
+            self.compare = self.grid
+
         self.rep = 0
-        self.params = None
-        self.objectivefunction = -1e308
-        self.bestrep = 0
+        self.parnames = parnames
+        self.parameters= len(parnames)
+        self.params_min = [np.nan]*self.parameters
+        self.params_max = [np.nan]*self.parameters
+        self.objectivefunction_min = 1e308
+        self.objectivefunction_max = -1e308
         self.starttime = time.time()
         self.last_print = time.time()
         
-        self.repetitions = None
+        self.repetitions = repetitions
+        self.stop = False
 
-    def __call__(self, rep, objectivefunction, params):
-        self.curparmeterset = params
+    def minimizer(self, objval, params):
+        if objval < self.objectivefunction_min:
+            self.objectivefunction_min = objval
+            self.params_min = list(params)
+
+    def maximizer(self, objval, params):
+        if objval > self.objectivefunction_max:
+            self.objectivefunction_max = objval
+            self.params_max = list(params)
+
+    def grid(self, objval, params):
+        if objval < self.objectivefunction_min:
+            self.objectivefunction_min = objval
+            self.params_min = list(params)
+        if objval > self.objectivefunction_max:
+            self.objectivefunction_max = objval
+            self.params_max = list(params)
+
+
+    def __call__(self, objectivefunction, params, block_print=False):
         self.rep+=1
-        if type(objectivefunction) == type([]):
-            if objectivefunction[0] > self.objectivefunction:
-                # Show only the first best objectivefunction when working with
-                # more than one objectivefunction
-                self.objectivefunction = objectivefunction[0]
-                self.params = params
-                self.bestrep = self.rep
+        if type(objectivefunction) == type([]): #TODO: change to iterable
+            self.compare(objectivefunction[0], params)
+        elif type(objectivefunction) == type(np.array([])):
+            pass
         else:
-            if objectivefunction > self.objectivefunction:
-                self.params = params
-                self.objectivefunction = objectivefunction
-                self.bestrep = self.rep
-        self.print_status()
+            self.compare(objectivefunction, params)
+
+
+        if self.rep == self.repetitions:
+            self.stop = True
+
+        if not block_print:
+            self.print_status()
 
     def print_status(self):
         # get str showing approximate timeleft to end of simulation in H, M, S
@@ -73,14 +108,59 @@ class _RunStatistic(object):
         if acttime - self.last_print >= 2:
             avg_time_per_run = (acttime - self.starttime) / (self.rep + 1)
             timestr = time.strftime("%H:%M:%S", time.gmtime(round(avg_time_per_run * (self.repetitions - (self.rep + 1)))))
-                    
-            text = '%i of %i (best like=%g) est. time remaining: %s' % (self.rep, self.repetitions,
-                                                                        self.objectivefunction, timestr)
+            if self.optimization_direction == 'minimize':
+                text = '%i of %i, minimal objective function=%g, time remaining: %s' % (
+                        self.rep, self.repetitions, self.objectivefunction_min, timestr)
+
+            if self.optimization_direction == 'maximize':
+                text = '%i of %i, maximal objective function=%g, time remaining: %s' % (
+                        self.rep, self.repetitions, self.objectivefunction_max, timestr)
+
+            if self.optimization_direction == 'grid':
+                text = '%i of %i, min objf=%g, max objf=%g, time remaining: %s' % (
+                        self.rep, self.repetitions, self.objectivefunction_min, self.objectivefunction_max, timestr)
+
             print(text)
             self.last_print = time.time()
-        
+
+    def print_status_final(self):
+        print('\n*** Final SPOTPY summary ***')
+        print('Total Duration: ' + str(round((time.time() - self.starttime), 2)) + ' seconds')
+        print('Total Repetitions:', self.rep)
+
+        if self.optimization_direction == 'minimize':
+            print('Minimal objective value: %g' % (self.objectivefunction_min))
+            print('Corresponding parameter setting:')
+            for i in range(self.parameters):
+                text = '%s: %g' % (self.parnames[i], self.params_min[i])
+                print(text)
+
+        if self.optimization_direction == 'maximize':
+            print('Maximal objective value: %g' % (self.objectivefunction_max))
+            print('Corresponding parameter setting:')
+            for i in range(self.parameters):
+                text = '%s: %g' % (self.parnames[i], self.params_max[i])
+                print(text)
+
+        if self.optimization_direction == 'grid':
+            print('Minimal objective value: %g' % (self.objectivefunction_min))
+            print('Corresponding parameter setting:')
+            for i in range(self.parameters):
+                text = '%s: %g' % (self.parnames[i], self.params_min[i])
+                print(text)
+
+            print('Maximal objective value: %g' % (self.objectivefunction_max))
+            print('Corresponding parameter setting:')
+            for i in range(self.parameters):
+                text = '%s: %g' % (self.parnames[i], self.params_max[i])
+                print(text)
+
+        print('******************************\n')
+
+
     def __repr__(self):
-        return 'Best objectivefunction: %g' % self.objectivefunction
+        return 'Min objectivefunction: %g \n Max objectivefunction: %g' % (
+                self.objectivefunction_min, self.objectivefunction_max)
 
 
 class _algorithm(object):
@@ -112,18 +192,12 @@ class _algorithm(object):
         seq: Sequentiel sampling (default): Normal iterations on one core of your cpu.
         mpc: Multi processing: Iterations on all available cores on your (single) pc
         mpi: Message Passing Interface: Parallel computing on high performance computing clusters, py4mpi needs to be installed
-    save_thresholde: float or list
+    save_threshold: float or list
         Compares the given value/list of values with return value/list of values from spot_setup.objectivefunction.
         If the objectivefunction value is higher, the results are saved in the database. If not they are ignored (saves storage).
     db_precision:np.float type
         set np.float16, np.float32 or np.float64 for rounding of floats in the output database
         Default is np.float16
-    alt_objfun: str or None, default: 'rmse'
-        alternative objectivefunction to be used for algorithm
-        * None: the objfun defined in spot_setup.objectivefunction is used
-        * any str: if str is found in spotpy.objectivefunctions, 
-            this objectivefunction is used, else falls back to None 
-            e.g.: 'log_p', 'rmse', 'bias', 'kge' etc.
     sim_timeout: float, int or None, default: None
         the defined model given in the spot_setup class can be controlled to break after 'sim_timeout' seconds if
         sim_timeout is not None.
@@ -132,35 +206,36 @@ class _algorithm(object):
         the algorithms uses the number in random_state as seed for numpy. This way stochastic processes can be reproduced.
     """
 
+    _unaccepted_parameter_types = (parameter.List, )
+
     def __init__(self, spot_setup, dbname=None, dbformat=None, dbinit=True,
-                 dbappend=False, parallel='seq', save_sim=True, alt_objfun=None,
-                 breakpoint=None, backup_every_rep=100, save_threshold=-np.inf,
-                 db_precision=np.float16, sim_timeout=None, random_state=None):
+                 dbappend=False, parallel='seq', save_sim=True, breakpoint=None,
+                 backup_every_rep=100, save_threshold=-np.inf, db_precision=np.float16,
+                 sim_timeout=None, random_state=None, optimization_direction='grid', algorithm_name=''):
+
         # Initialize the user defined setup class
         self.setup = spot_setup
-        self.model = self.setup.simulation
-        # Philipp: Changed from Tobi's version, now we are using both new class defined parameters
-        # as well as the parameters function. The new method get_parameters
-        # can deal with a missing parameters function
-        #
-        # For me (Philipp) it is totally unclear why all the samplers should call this function
-        # again and again instead of
-        # TODO: just storing a definite list of parameter objects here
+        param_info = parameter.get_parameters_array(self.setup, unaccepted_parameter_types=self._unaccepted_parameter_types)
+        self.all_params = param_info['random']
+        self.constant_positions = parameter.get_constant_indices(spot_setup)
+        if self.constant_positions:
+            self.non_constant_positions = []
+            for i, val in enumerate(self.all_params):
+                if self.all_params[i] not in self.constant_positions:
+                    self.non_constant_positions.append(i)
+        else: 
+            self.non_constant_positions = np.arange(0,len(self.all_params))
         self.parameter = self.get_parameters
-        self.parnames = self.parameter()['name']
-
+        self.parnames = param_info['name']
+        self.algorithm_name = algorithm_name
         # Create a type to hold the parameter values using a namedtuple
-        self.partype = parameter.get_namedtuple_from_paramnames(
-            self.setup, self.parnames)
+        self.partype = parameter.ParameterSet(param_info)
 
-        # use alt_objfun if alt_objfun is defined in objectivefunctions,
-        # else self.setup.objectivefunction
-        self.objectivefunction = getattr(
-            objectivefunctions, alt_objfun or '', None) or self.setup.objectivefunction
         self.evaluation = self.setup.evaluation()
         self.save_sim = save_sim
+        self.optimization_direction = optimization_direction
         self.dbname = dbname or 'customDb'
-        self.dbformat = dbformat or 'custom'
+        self.dbformat = dbformat or 'ram'
         self.db_precision = db_precision
         self.breakpoint = breakpoint
         self.backup_every_rep = backup_every_rep
@@ -171,9 +246,9 @@ class _algorithm(object):
         self.dbappend = dbappend
         
         # Set the random state
-        if random_state is None:
+        if random_state is None: #ToDo: Have to discuss if these 3 lines are neccessary.
             random_state = np.random.randint(low=0, high=2**30)
-        np.random.seed(random_state)
+        np.random.seed(random_state) 
 
         # If value is not None a timeout will set so that the simulation will break after sim_timeout seconds without return a value
         self.sim_timeout = sim_timeout
@@ -181,8 +256,11 @@ class _algorithm(object):
 
         if breakpoint == 'read' or breakpoint == 'readandwrite':
             print('Reading backupfile')
+            try:
+                open(self.dbname+'.break')
+            except FileNotFoundError:
+                print('Backupfile not found')
             self.dbappend = True
-            self.breakdata = self.read_breakdata(self.dbname)
 
         # Now a repeater (ForEach-object) is loaded
         # A repeater is a convinent wrapper to repeat tasks
@@ -214,12 +292,12 @@ class _algorithm(object):
         # the normal work on the chains
         self.repeat = ForEach(self.simulate)
 
-        # In MPI, this command will do nothing on the master process
-        # but the worker processes are going to wait for jobs.
-        # Hence the workers will only receive parameters for the
-        # simulate function, new calculation phases and the termination
-        self.repeat.start()
-        self.status = _RunStatistic()
+
+        # method "save" needs to know whether objective function result is list or float, default is float
+        self.like_struct_typ = type(1.1)
+
+
+
 
     def __str__(self):
         return '{type}({mtype}())->{dbname}'.format(
@@ -234,24 +312,26 @@ class _algorithm(object):
         """
         Returns the parameter array from the setup
         """
-        return parameter.get_parameters_array(self.setup)
+        pars = parameter.get_parameters_array(self.setup)
+        return pars[self.non_constant_positions]
 
     def set_repetiton(self, repetitions):
-        self.status.repetitions = repetitions
-        
+        self.status = _RunStatistic(repetitions, self.algorithm_name,
+                                    self.optimization_direction, self.parnames)
+        # In MPI, this command will do nothing on the master process
+        # but the worker processes are going to wait for jobs.
+        # Hence the workers will only receive parameters for the
+        # simulate function, new calculation phases and the termination
+        self.repeat.start()
+
     def final_call(self):
         self.repeat.terminate()
         try:
             self.datawriter.finalize()
         except AttributeError:  # Happens if no database was assigned
             pass
-        print('End of sampling')
-        text = 'Best run at %i of %i (best like=%g) with parameter set:' % (
-            self.status.bestrep, self.status.repetitions, self.status.objectivefunction)
-        print(text)
-        print(self.status.params)
-        text = 'Duration:' + str(round((time.time() - self.status.starttime), 2)) + ' s'
-        print(text)
+        self.status.print_status_final()
+
 
     def _init_database(self, like, randompar, simulations):
         if self.dbinit:
@@ -265,56 +345,79 @@ class _algorithm(object):
 
             self.dbinit = False
 
+
+    def __is_list_type(self, data):
+        if type(data) == type:
+            return data == list or data == type(np.array([]))
+        else:
+            return type(data) == list or type(data) == type(np.array([]))
+
     def save(self, like, randompar, simulations, chains=1):
         # Initialize the database if no run was performed so far
         self._init_database(like, randompar, simulations)
 
-        #try if like is a list of values compare it with save threshold setting
-        try:
+        if self.__is_list_type(self.like_struct_typ) and self.__is_list_type(self.save_threshold):
             if all(i > j for i, j in zip(like, self.save_threshold)): #Compares list/list
                 self.datawriter.save(like, randompar, simulations, chains=chains)
-        #If like value is not a iterable, it is assumed to be a float
-        except TypeError: # This is also used if not threshold was set
-            try:
-                if like>self.save_threshold: #Compares float/float
-                    self.datawriter.save(like, randompar, simulations, chains=chains)
-            except TypeError:# float/list would result in an error, because it does not make sense
-                if like[0]>self.save_threshold: #Compares list/float
-                    self.datawriter.save(like, randompar, simulations, chains=chains)
+        if not self.__is_list_type(self.like_struct_typ) and not self.__is_list_type(self.save_threshold):
+            if like>self.save_threshold: #Compares float/float
+                self.datawriter.save(like, randompar, simulations, chains=chains)
+        if self.__is_list_type(self.like_struct_typ) and not self.__is_list_type(self.save_threshold):
+            if like[0]>self.save_threshold: #Compares list/float
+                self.datawriter.save(like, randompar, simulations, chains=chains)
+        if not self.__is_list_type(self.like_struct_typ) and self.__is_list_type(self.save_threshold): #Compares float/list
+            if (like > self.save_threshold).all:
+                self.datawriter.save(like, randompar, simulations, chains=chains)
 
     def read_breakdata(self, dbname):
         ''' Read data from a pickle file if a breakpoint is set.
             Reason: In case of incomplete optimizations, old data can be restored. '''
         import pickle
         with open(dbname+'.break', 'rb') as breakfile:
-            return pickle.load(breakfile)
+            work,backuptime,repos,obmin,obmax=pickle.load(breakfile)
+            self.status.starttime=self.status.starttime-backuptime
+            self.status.rep=repos
+            self.status.objectivefunction_min=obmin
+            self.status.objectivefunction_max=obmax
+            return work
 
     def write_breakdata(self, dbname, work):
         ''' Write data to a pickle file if a breakpoint has been set.'''
         import pickle
+        work=(work,self.status.last_print-self.status.starttime,self.status.rep,self.status.objectivefunction_min,self.status.objectivefunction_max)
         with open(str(dbname)+'.break', 'wb') as breakfile:
             pickle.dump(work, breakfile)
 
     def getdata(self):
         return self.datawriter.getdata()
 
-    def postprocessing(self, rep, randompar, simulation, chains=1, save=True, negativlike=False):
-        like = self.getfitness(simulation=simulation, params=randompar)
+    def update_params(self, params):
+        #Add potential Constant parameters
+        self.all_params[self.non_constant_positions] = params
+        return self.all_params
+            
+    
+    def postprocessing(self, rep, params, simulation, chains=1, save_run=True, negativlike=False, block_print=False): # TODO: rep not necessaray
+    
+        params = self.update_params(params)
+        if negativlike is True:
+            like = -self.getfitness(simulation=simulation, params=params)
+        else:
+            like = self.getfitness(simulation=simulation, params=params)
+
         # Save everything in the database, if save is True
         # This is needed as some algorithms just want to know the fitness,
         # before they actually save the run in a database (e.g. sce-ua)
-        if save is True:
-            if negativlike is True:
-                self.save(-like, randompar, simulations=simulation, chains=chains)              
-                self.status(rep, -like, randompar)
-            else:
-                self.save(like, randompar, simulations=simulation, chains=chains)
-                self.status(rep, like, randompar)
+
+        self.status(like,params,block_print=block_print)
+        
+        if save_run is True and simulation is not None:
+            self.save(like, params, simulations=simulation, chains=chains)
         if type(like)==type([]):
             return like[0]
         else:        
             return like
-    
+
     
     def getfitness(self, simulation, params):
         """
@@ -322,11 +425,11 @@ class _algorithm(object):
         """
         try:
             #print('Using parameters in fitness function')
-            return self.objectivefunction(evaluation=self.evaluation, simulation=simulation, params = (params,self.parnames))
+            return self.setup.objectivefunction(evaluation=self.evaluation, simulation=simulation, params = (params,self.parnames))
 
         except TypeError: # Happens if the user does not allow to pass parameter in the spot_setup.objectivefunction
             #print('Not using parameters in fitness function')            
-            return self.objectivefunction(evaluation=self.evaluation, simulation=simulation)
+            return self.setup.objectivefunction(evaluation=self.evaluation, simulation=simulation)
     
     def simulate(self, id_params_tuple):
         """This is a simple wrapper of the model, returning the result together with
@@ -334,16 +437,18 @@ class _algorithm(object):
         can mix up the ordering of runs
         """
         id, params = id_params_tuple
+        self.all_params[self.non_constant_positions] = params #TODO: List parameters are not updated if not accepted for the algorithm, we may have to warn/error if list is given
+        all_params = self.all_params
 
         # we need a layer to fetch returned data from a threaded process into a queue.
-        def model_layer(q,params):
+        def model_layer(q,all_params):
             # Call self.model with a namedtuple instead of another sequence
-            q.put(self.model(self.partype(*params)))
+            q.put(self.setup.simulation(self.partype(*all_params)))
 
         # starting a queue, where in python2.7 this is a multiprocessing class and can cause errors because of
         # incompability which the main thread. Therefore only for older Python version a workaround follows
         que = Queue()
-        sim_thread = threading.Thread(target=model_layer, args=(que, params))
+        sim_thread = threading.Thread(target=model_layer, args=(que, all_params))
         sim_thread.daemon = True
         sim_thread.start()
 
@@ -353,8 +458,8 @@ class _algorithm(object):
         sim_thread.join(self.sim_timeout)
 
         # If no result from the thread is given, i.e. the thread was killed from the watcher the default result is
-        # '[nan]' otherwise get the result from the thread
-        model_result = [np.NAN]
+        # '[nan]' and will not be saved. Otherwise get the result from the thread
+        model_result = None
         if not que.empty():
             model_result = que.get()
         return id, params, model_result
